/*****************************************************************************/
/*****************************************************************************/
/******************            main.c                  ***********************/
/*****************************************************************************/
/*****************************************************************************/

/*----------------------------------------------------------------------------
Changes to version 1.1
 
August 2013, Twenty one years later... Generation of a truly random path
August 2006. Hole_effect variogram included by Peter Salomon as case 5.

Changes to version 1.0

August 1992. Undo the change mentioned below.
August 1992. Rotate the axes to classify the points in octants, so
           that the anisotropy axes are not aligned to the search axes.
August 1992. The parameter dimension has been changed to maximum_number_
           of_data and if max_per_octant==0 then no octant search is
           considered

July 1992. add number of dimensions to the input data file. Not anymore
           need to define TWO_D.
July 1992. Add findout_rotation. There is not anymore need to set the search
           rotation flag.
July 1992. Change of the output file format. Right now the new write_out
           will write out a file in TOOLBOX/GEOEAS format, that is, a
           title, then the number of variables simulated, then the
           names of the simulated variables and then by columns all
           node values which are being written layer by layer, each
           layer written by rows. First node written the one with
           coordinates origin.x origin.y origin.z.
July 1992. Many realizations are generated, not in the most efficient
           manner but it works.
-----------------------------------------------------------------------------*/

#include "ghead12.h"

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - -    main()   - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
int main()
/*  THREE DIMENSIONAL MULTIPLE GAUSSIAN CONDITIONAL CO-SIMULATION PROGRAM
    ---------------------------------------------------------------------

        J. Jaime Gomez-Hernandez

        Department of Applied Earth Sciences
        Stanford, CA 94305


To refer to this program use:

Gomez-Hernandez, J. Jaime (1991) A Stochastic Approach to the Simulation of
Block Conductivity values Conditioned Upon Data Measured at a Smaller Support,
Ph. D. Thesis, Stanford University, CA. Available from Ann Harbor microfilms,
Michigan.

Gomez-Hernandez, J. Jaime and A. G. Journel (1993) Joint Sequential Simulation
of MultiGaussian Fields, in Troia '92, Amilcar Soares (Ed.), volume 1,
pages 85-94, Kluwer Academic Publishers.

Summary:
    -This program produces conditional multigaussian simulations of several
     variables jointly multigaussian. The purpose of the main is
     to organize the main subroutines needed for the simulation. The purpose
     of each subroutine is outlined at the beginning of the subroutine.

     The program performs a conditional sequential gaussian simlation on a
     rectangular grid of several variables.
     The input data required is divided in three groups, the
     geometry data, the variogram data and the conditioning data. These data is
     input from three different data files, the name of which is #define'd in
     the header file "gcosim3d.h". The program outputs three different kinds of
     information: error messages, debugging information and simulation per se.

------------- INPUT DATA FILES ---------------------------

    -The file GEOMETRY contains the following information
    record 1: seed, nsim, an integer used as seed for the
              randon number generator, number of simulations..
    record 2: delta.x, delta.y, delta.z, spacing in the x,y,z directions
            of the simulation grid
    record 3: origin.x, origin.y, origin.z, coordinates of simulation grid
            point (1,1,1)
    record 4: n_nodes.x, n_nodes.y, n_nodes.z, the number of nodes in the
            x,y,z directions of the simulation grid
    ***note*** for debugging and testing purposes, it is useful to be able
            to limit the simulation to some particular sub-area of
            the entire grid. The following parameters specify a
            rectangular area for the actual simulation. If a
            complete simulation is required, then from(x,y,z)
            should be set to 1 and to(x,y,z) to the total number
            of nodes.
    record 5: from.x, from.y, from.z, x,y and z indices of the beginning
            of the simulation sub-area
    record 6: to.x, to.y, to.z, x,y,z indices of the end of the simulation
            sub-area
    record 7: radius.x, radius.y, radius.z, search ellipsoid radii.
    record 8,9 and 10: direction cosines of the rectangular system of
            search axes. Same for all structures and all variograms.
    record 11: max_per_octant, , max_data, maximum number of points per
            octant to be retained for kriging and overall maximum of
            data in total. (If max_per_octant==0, no octant search
            is implemented).
    record 12: dbg, debugging flag, generally set to 0, if set to 3 or
            higher, a large amount of information is output, this
            option should be used only if the simulation subarea
            is very small.

    -The file VARIOGRAM contains the following information
    record 1: nvar, number of variables in the data file
    record 2: nvar_sim, number of variables that are going to be simulated
              (in some cases, data is available for two variables but only
               one is to be simulated ---> nvar_sim <= nvar)
    record 3: simulation_var[1..nvar_sim] an array of indices indicating which
              variables are to be simulated
    record 4: expected_value[1..nvar], the expected value of each of the
              variables in the input data file. It can be split in several lines

    ---------The following records are repeated for each one of the variograms
             and cross-variograms. The order in which these are input
             is given by the rows of the lower triangular
             half of the variance-covariance matrix, that is,
             first   variogram of variable 1
             second  cross-variogram between variables 1 and 2
             third   variogram of variable 2
             fourth  cross-variogram between variables 1 and 3
             fifth   cross-variogram between variables 2 and 3
             sixth   variogram of variable 3
             and so on.
    record 5: nugget, of the variogram or cross-variogram
    record 6: cmax, an upper bound of the maximum value that the variogram
            can reach within the limits of the search neighborhood.
            To be used with power variograms only, although a value must
            be input for all variogram models
    record 7: num_struct, number of nested structures in the variogram
    record 8: type, of each nested structure
            1: spherical
            2: exponential
            3: gaussian
            4: power
            5: hole_effect
	        6: gneiting (with some fixed parameters in the function defintion)
    record 9: sill, of each structure (scaling coefficient for the power
            model)
    record 10: ranges of each structure in the x direction
            after rotation of the cartesian axes
    record 11: ranges of each structure in the y direction
            after rotation of the cartesian axes
    record 12: ranges of each structure in the z direction
            after rotation of the cartesian axes
    record 13,14,15: direction cosines of the rectangular system of
            anisotropy axes.

    Records 5 to 15 are repeated until a total of (nvar(nvar+1)/2) variograms
    and cross-variograms are input

    -The file DATA contains the following information
    record 1: ndata, number of hard conditioning data of all types
    record 2 to record ndata+1: x,y,z,variable_index,value: coordinates, an
             index (<= nvar) indicating to which variable this value corresponds,
             and value.

------------- OUPUT DATA FILES ---------------------------

    -Error messages are displayed in the console and generally mean the
    termination of the execution. Also there is feedback information
    sent to the console to keep the user informed of the progress of
    the execution.

    -The file DBG will receive information such as the input data from
     all files and different parameters that are computed during execution
     and that are of relevance to the final result. If the debug flag is
     set (in record 13 of GEOMETRY) to a value greater than 2 almost
     every significative intermediate result is written out.

    -The file OUTPUT will receive the final simulation. The way this file
    should be written depends very much in the planned use of the
    results. The output is controlled by the subroutine 'write_out_results'
    that should be rewritten for each specific necessity. An example
    of such subroutine is given.
*/

{
    FILE            *fgeom, *fvario, *fdata, *fout, *fdebug;
    float           *expected_value, *value, ****cova_table;
    pt              delta, origin, *coord;
    ipt             from, to,n_nodes,srch_dim;
    int             max_per_octant, dbg, ndata, nvar,nvar_sim,*simulation_var
                    ,ntemp,ncov,acov,**lookup_table,*index,nsim,i,max_data;
    search          srch;
    variogram_model *variogram;
    float           ****node;
    short           ***node_flag;
    long int        seed;
    template        tmp;

    /* DICTIONARY OF VARIABLES

    acov            the index of one of the covariance tables
    *coord          vector of 'pt's with the coordinates of the
                        conditioning data
    ****cova_table  tensor of 4 dimensions with covariances. Element
                        cova_table[i][dx][dy][dz] contains the covariance
                        with index 'i' between two nodes
                        separated by 'dx', 'dy' and 'dz' grid nodes.
    dbg             debugging flag
    delta           a 'pt' that contain the grid spacing
    *fdata          a pointer to file DATA
    *fdebug         a pointer to file DBG
    *fgeom          a pointer to file GEOMETRY
    *fout           a pointer to file OUTPUT
    from            a 'ipt' with the indices of the first node of the
                        simulation sub-area
    *fvario         a pointer to file VARIOGRAM
    *index          a vector of integers, index[i] indicates which variable
                    is value[i] in the input data stream
    **lookup_table  is a double entry table that relates the indices of the
                    two variables being considered and the position of the
                    corresponding auto- or cross-covariance in the covariance
                    tables
    max_data        maximum number of data values to be retained within
                       the search neighborhood for kriging
    max_per_octant  maximum number of points per octant to be retained for
                        kriging. If it is equal to 0, no octant search
                        is implemented
    ncov            total number of variograms and cross-variograms
    ndata           number of conditioning data
    nsim            number of simulations
    nvar            number of variables in total
    nvar_sim        number of variables to be simulated
    n_nodes         a 'ipt' with the number of nodes in each direction
    ****node        a tensor of dimension 4 with the values of the variables
                        for each node of the simulation grid. node[iv][ix][iy][iz]
                        correspond to the value of variable [iv] at
                        node with grid coordinates (ix,iy,iz). By convention
                        if node[iv][ix][iy][iz] == UNKNOWN, the value of
                        variable iv at (ix,iy,iz) has not been simulted yet.
                        The value of UNKNOWN is #define'd in 'gcosim3d.h'
    ***node_flag    a tensor of dimension 3 with a flag indicating whether
                    the corresponding node has been simulated.
                       node_flag[ix][iy][iz] == -1  node has not been simulated
                       node_flag[ix][iy][iz] == 0   node is partially simulated
                       node_flag[ix][iy][iz] == 1   node completely simulated
    ntemp           number of points in search template
    origin          a 'pt' with the coordinates of node[1][1][1]
    seed            random generator seed value
    *simulation_var an array of indices indicating which variables are to be
                    simulated
    srch            search neighborhood
    srch_dim        dimensions of the smallest parallelepiped that contains
                        the search neighborhood
    *expected_value      vector of expected values for each variable
    tmp             search template
    to              a 'ipt' with the indices of the last point of the
                    ______    simulation sub-area
    *value          a vector with the data value of the conditioning data
    *variogram      a vector with the variogram and cross-variograms of the
                         variables
    */

                                                               /* open files */

    if ( (fgeom=fopen(GEOMETRY,"r"))==NULL )
        error("cannot open file GEOMETRY");
    if ( (fvario=fopen(VARIOGRAM,"r"))==NULL )
        error("cannot open file VARIOGRAM");
    if ( (fdata=fopen(DATA,"r"))==NULL )
        error("cannot open file DATA");
    if ( (fout=fopen(OUTPUT,"w"))==NULL )
        error("cannot open file OUTPUT");
    if ( (fdebug=fopen(DBG,"w"))==NULL )
        error("cannot open file DBG");

    print_header(fdebug);
                                                             /* read in data */

    read_in_geometry(fgeom,fdebug,&seed,&nsim,&max_data,&delta,&origin,&n_nodes,
            &from,&to,&srch,&max_per_octant,&dbg);
    fclose(fgeom);

    nvar=read_in_variograms(fvario,fdebug,&nvar_sim,&simulation_var,
                            &expected_value,&variogram);
    ncov=nvar*(nvar+1)/2;
    fclose(fvario);

    ndata=read_in_conditioning_data(fdata,fdebug,&coord,&index,&value);
    fclose(fdata);

                                                          /* allocate_memory */

    node=allocate_nodes(nvar,n_nodes,&node_flag);
    cova_table=allocate_covariance_tables(ncov,srch,&srch_dim,delta,
                    &lookup_table);
    ntemp=allocate_template(fdebug,&tmp,srch_dim,srch,delta);

                                          /* initialize variables and arrays */
    initialize_covariance_tables(fdebug,dbg,ncov,
            srch_dim,delta,cova_table,variogram,lookup_table);

                                         /* the templates are initialized using
                                            the covariance table of the first
                                            variable to be simulated */

    acov=lookup_table[simulation_var[1]][simulation_var[1]];
    initialize_template(fdebug,dbg,&tmp,srch_dim,srch,delta,
            cova_table[acov]);

                                                  /* some consistency checks */

    a_few_checks(ndata,ntemp,ncov,nvar,nvar_sim,n_nodes,from,to,delta);

                                             /* seed random number generator */

    if(seed>0) seed=(-seed); rand0(&seed);

    for (i=1;i<=nsim;i++) {

    initialize_nodes(fdebug,n_nodes,node,node_flag,ndata,coord,index,value,
            origin,delta,nvar,i);

                                                        /* simulation starts */

    do_simulation(fdebug,n_nodes,from,to,max_per_octant,nvar,nvar_sim,
            simulation_var,expected_value,node,node_flag,
            ntemp,&tmp,&seed,cova_table,lookup_table,dbg,max_data,i);

                                                                /* print out */

    write_out_results(fout,from,to,node,nvar_sim,simulation_var,i,nsim);

/*  fprintf(fout,"\n"); */

    }
    exit(0); /* normal termination */
}


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - -     write_out_results()   - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void
write_out_results(fout,from,to,node,nvar_sim,simulation_var,isim,nsim)
FILE *fout;
ipt from,to;
float ****node;
int nvar_sim,*simulation_var,isim,nsim;
{

/* This is an example of a subroutine to output the simulation results. In this
   case each of the output records contains as many columns as variables
   have been simulated   */

    int i,ix,iy,iz,iv;

    /* if this is the first simulation write header to output file */

    if(isim==1) {
       fprintf(fout,"Output from GCOSIM3D, %d simulations of the subarea\n",nsim);
       fprintf(fout,"%d\n",nvar_sim);
       for(i=1;i<=nvar_sim;i++)
           fprintf(fout,"Variable number %d\n",simulation_var[i]);
    }


    /* output by layers, then by rows, then by columns */
    /* all variables in the same output record */

    for(iz=from.z;iz<=to.z;iz++) {
    for(iy=from.y;iy<=to.y;iy++) {
    for(ix=from.x;ix<=to.x;ix++) {
          for(i=1;i<=nvar_sim;i++) {
              iv=simulation_var[i];
              fprintf(fout,"%7.3f ",node[iv][ix][iy][iz]);
          }
          fprintf(fout,"\n");
    }
    }
    }
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - rotate()  - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

pt rotate(vect,dir_cos)
pt vect;
float   dir_cos[3][3];
{
/* this function returns the vector 'vect' after rotation according to the
   direction cosines in the 3 by 3 matrix dir_cos. The rotated vector
   returns in the same vector vect */

    pt  rotated;

    rotated.x=vect.x*dir_cos[0][0]+
              vect.y*dir_cos[0][1]+
              vect.z*dir_cos[0][2];

    rotated.y=vect.x*dir_cos[1][0]+
              vect.y*dir_cos[1][1]+
              vect.z*dir_cos[1][2];

    rotated.z=vect.x*dir_cos[2][0]+
              vect.y*dir_cos[2][1]+
              vect.z*dir_cos[2][2];

    return rotated;
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - hole_effect() - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float hole_effect(float h,float sill,float a)
{
/* hole_effect isotropic variogram

    double exp();  */

    h=h/a;
    return (sill*(1.0-cos((double)h*3.1415927)));
}


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - exponential() - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float exponential(float h,float sill,float a)
{
/* exponential isotropic variogram

    double exp();  */

    h=h/a;
    return (sill*(1.0-exp((double)-h)));
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - spherical() - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float spherical(float h,float sill,float a)
{
/* spherical isotropic variogram */

    h=h/a;
    return (h<1.0 ? sill*h*(1.5-0.5*h*h) : sill);
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - gaussian()  - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float gaussian(float h,float sill,float a)
{
/* gaussian isotropic variogram

    double exp();  */

    h=h/a;
    return (sill*(1.0-exp((double)-h*h)));
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - linear()  - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float linear(float h,float slope,float power)
{
/* linear isotropic variogram */

    double pow();

    return slope*pow((double) h,(double) power);
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - modulus() - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float modulus(vect,anis)
pt vect,anis;
{
/* apply anisotropy ratios and compute the modulus of a vector */

    float h;

    h=vect.x*vect.x*anis.x*anis.x+
      vect.y*vect.y*anis.y*anis.y+
      vect.z*vect.z*anis.z*anis.z;
    h=sqrt((double) h);

    return h;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - cova()  - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float cova(point1,point2,var_model)
pt point1,point2;
variogram_model *var_model;
{
/* this function returns the covariance between points 'point1' and 'point2'
   according to the variogram model var_model. The covariance is obtained from
   the variogram value is substracted */

    float cov;
    int i;

    cov=var_model->nugget;
    for(i=0;i<var_model->num_struct;i++) {
        switch (var_model->type[i]) {
            case 1:
            case 2:
            case 3:
			case 5:
                    cov+=var_model->sill[i];
                    break;
            case 4: cov+=var_model->cmax;
        }
    }
    cov=cov-gamma0(point1,point2,var_model);
    return cov;
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - gamma0()  - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float gamma0(point1,point2,var_model)
pt point1, point2;
variogram_model *var_model;
{
/* this function returns the variogram value between points 'point1' and
   'point2' according to the variogram model var_model. The four, most common
   variogram models are considered. */

    float  value,h;
    int    i;
    pt     delta;

    /* obtain separation vector between point1 and point2
       and store it in point1 */

    delta.x=point1.x-point2.x;
    delta.y=point1.y-point2.y;
    delta.z=point1.z-point2.z;

    delta=rotate(delta,var_model->dir_cosines);

    h=delta.x*delta.x+delta.y*delta.y+delta.z*delta.z;
    if(h<=1e-10) return 0.;

    value=var_model->nugget;                           /* add nugget effect */

    for(i=0;i<var_model->num_struct;i++) {    /* loop through all structures */

    /* first obtain the moudulus of the rotated vector after applying
       the anisotropy corrections */

    h=modulus(delta,var_model->anis[i]);

    if (h<=1.e-20F) continue;  /* for very small distances the variogram is 0 */


        switch(var_model->type[i])  {
            case 1: value+=spherical(h,var_model->sill[i],
                                              var_model->a[i]);
                    break;

            case 2: value+=exponential(h,var_model->sill[i],
                                              var_model->a[i]);
                    break;

            case 3: value+=gaussian(h,var_model->sill[i],
                                              var_model->a[i]);
                    break;

            case 4: value+=linear(h,var_model->sill[i],
                                              var_model->a[i]);
                    break;

            case 5: value+=hole_effect(h,var_model->sill[i],
                                              var_model->a[i]);
                    break;

            default: error("Undefined variogram type");
        } /* end switch */
    }     /* end for    */

    return value;
}





/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - allocate_nodes()  - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float ****allocate_nodes(nvar,n,node_flag_pointer)
ipt n;
short ****node_flag_pointer;
int nvar;
{
/* This subroutine returns a pointer to a tensor of dimensions [1..nvar]
   [1..nx][1..ny][1..nz] that will store the values of the variables for each
   node of the simulation grid. */

    int     i,j,k;
    float   ****t;
    short   ***node_flag;


    printf("Allocating nodes\n\n");

                                    /* allocate pointers to variables */

    t=(float ****) malloc((size_t) (nvar)*sizeof(float ***));
    if(!t) error("allocation failure 1 in allocate_nodes()");

    t--;
    								/* allocate a row of pointers to columns */
    for (i=1;i<=nvar;i++) {
        t[i]=(float ***) malloc((size_t) (n.x)*sizeof(float **));
        if(!t[i]) error("allocation failure 2 in allocate_nodes()");
        t[i]--;
    }
               /* allocate columns of pointers and set pointers to them */

    for (i=1;i<=nvar;i++)
        for(j=1;j<=n.x;j++) {
            t[i][j]=(float **) malloc((size_t)(n.y)*sizeof(float *));
            if(!t[i][j]) error("allocation failure 3 in allocate_nodes()");
            t[i][j]--;
        }

                                 /* allocate levels and set pointers to them */

    for (i=1;i<=nvar;i++)
        for (j=1;j<=n.x;j++)
            for(k=1;k<=n.y;k++) {
               t[i][j][k]=(float *) malloc((size_t)(n.z)*sizeof(float));
               if(!t[i][j][k]) error("allocation failure 4 in allocate_nodes()");
               t[i][j][k]--;
            }
    								/* allocate a row of pointers to columns */

    node_flag=(short ***) malloc((size_t) (n.x)*sizeof(short **));
    if(!node_flag) error("allocation failure 5 in allocate_nodes()");

               /* allocate columns of pointers and set pointers to them */
    node_flag--;

    for (i=1;i<=n.x;i++) {
        node_flag[i]=(short **) malloc((size_t) (n.y)*sizeof(short *));
        if(!node_flag[i]) error("allocation failure 6 in allocate_nodes()");
        node_flag[i]--;
    }

                                 /* allocate levels and set pointers to them */

    for (i=1;i<=n.x;i++)
        for(j=1;j<=n.y;j++) {
            node_flag[i][j]=(short *) malloc((size_t)(n.z)*sizeof(short));
            if(!node_flag[i][j]) error("allocation failure 7 in allocate_nodes()");
            node_flag[i][j]--;
        }

    *node_flag_pointer=node_flag;

    return t;
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - -  allocate_covariance_tables() - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float ****allocate_covariance_tables(ncov,srch,dim,delta,lookup_table_pointer)
search  srch;
pt      delta;
int     ncov;
ipt     *dim;
int     ***lookup_table_pointer;
{

/* This subroutine returns the pointer to a tensor of dimension 4
   that will contain the covariance tables. The dimensions of the covariance
   tables is determined by the smallest parallelepiped that can contain
   the search neighborhood 'srch'. This dimensions are computed according
   to the rotation flag (refer to the 'isim3d.h' for description of the
   structure 'search'. An ergodic covariance is assumed, therfore
   cova(dx,dy,dz)= cova(-dx,-dy,-dz) so that it needs to be computed only
   for positive values of dx */

    int i,**lookup_table;
    float ****cova_table;

                         /* optimize the dimensions of the covariance tables */

    printf("Allocating covariance tables\n\n");

    delta.x=fabs((double)delta.x);
    delta.y=fabs((double)delta.y);
    delta.z=fabs((double)delta.z);

    switch (srch.rotation) {
    case 0: dim->x=NINT(srch.radius.x/delta.x);             /* no rotation */
            dim->y=NINT(srch.radius.y/delta.y);
            dim->z=NINT(srch.radius.z/delta.z);
            break;
    case 1: dim->x=NINT(srch.radius.x/delta.x);       /* rotation around x */
            dim->y=NINT(MAX(srch.radius.y,srch.radius.z)/delta.y);
            dim->z=NINT(MAX(srch.radius.y,srch.radius.z)/delta.z);
            break;
    case 2: dim->x=NINT(MAX(srch.radius.x,srch.radius.z)/delta.x);
            dim->y=NINT(srch.radius.y/delta.y);       /* rotation around y */
            dim->z=NINT(MAX(srch.radius.x,srch.radius.z)/delta.z);
            break;
    case 3: dim->x=NINT(MAX(srch.radius.x,srch.radius.y)/delta.x);
            dim->y=NINT(MAX(srch.radius.x,srch.radius.y)/delta.y);
            dim->z=NINT(srch.radius.z/delta.z);       /* rotation around z */
            break;
    case 4: dim->x=NINT(MAX(srch.radius.x,MAX(srch.radius.y,srch.radius.z))
                    /delta.x);
            dim->y=NINT(MAX(srch.radius.x,MAX(srch.radius.y,srch.radius.z))
                    /delta.y);
            dim->z=NINT(MAX(srch.radius.x,MAX(srch.radius.y,srch.radius.z))
                    /delta.z);
            break;
    default: error("rotation flag out of limits");
    }

                                                     /* allocation of tables */
    cova_table=(float ****)malloc(ncov*sizeof(float ***));
    if(!cova_table) error("Covariance tables allocation failure");
    cova_table--;
    for(i=1;i<=ncov;i++)
        cova_table[i]=tensor(0,2*dim->x,-2*dim->y,2*dim->y,-2*dim->z,2*dim->z);

    lookup_table=imatrix(1,ncov,1,ncov);

    *lookup_table_pointer=lookup_table;
    return cova_table;
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - allocate_template() - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

int allocate_template(fperr,tmp,dim,srch,delta)
FILE        *fperr;
template    *tmp;
ipt         dim;
search      srch;
pt          delta;
{

/* This subroutine allocates the members of the 'tmp' structure. The 'tmp'
   structure contains all the information associated with the search template,
   that is, relative coordinates of all the points within a search
   neighborhood and distance to the point around which the search is taking
   place, as well as the octant to which each point belongs. There is also
   an 'index' member to retrieve the template points in order, from the
   closest to the farthest */

    int         ix,iy,iz,i,ntemp;
    pt          h;
    float       norm_h;

                                        /* find out the size of the template */

    printf("Allocating template\n\n");

    ntemp=0;
    for(ix= -dim.x;ix<=dim.x;ix++)
        for(iy= -dim.y;iy<=dim.y;iy++)
            for(iz= -dim.z;iz<=dim.z;iz++) {
                h.x=ix*delta.x;
                h.y=iy*delta.y;
                h.z=iz*delta.z;
                h=rotate(h,srch.cosines);
                h.x/=srch.radius.x;
                h.y/=srch.radius.y;
                h.z/=srch.radius.z;
                norm_h=h.x*h.x+h.y*h.y+h.z*h.z;
                ntemp+=(norm_h < 1 ) ? 1 : 0;
            }

                                                        /* allocate template */

    tmp->irx=ivector(1,ntemp);
    tmp->iry=ivector(1,ntemp);
    tmp->irz=ivector(1,ntemp);
    tmp->octant=ivector(1,ntemp);
    tmp->index=ivector(1,ntemp);
    tmp->distance=vector(1,ntemp);

    fprintf(fperr,"\n\nNumber of points in template -------> %d\n\n",ntemp);

    return ntemp;
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - initialize_nodes()  - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/


void initialize_nodes(fperr,n_nodes,node,node_flag,ndata,coord,index,value,
                        origin,delta,nvar,isim)
FILE        *fperr;
ipt         n_nodes;
float       ****node;
short       ***node_flag;
int         *index,isim;
int         ndata,nvar;
float       value[];
pt          coord[],delta,origin;
{

/* This subroutine assigns each conditioning data to the closest node in
    the simulation grid. It also prints warning messages if two points
    are assigned to the same node or if a point is outside of the simulation
    area */

    int ix,iy,iz,i;   /* ix,iy,iz are the grid indices assigned to the datum */
    int iv;         /* the variable index */

    printf("Initializing nodes--");
                                              /* set simulation flag to -1 */
    for (ix=1;ix<=n_nodes.x;ix++)
        for (iy=1;iy<=n_nodes.y;iy++)
            for (iz=1;iz<=n_nodes.z;iz++) {
                node_flag[ix][iy][iz]=-1;
                for(iv=1;iv<=nvar;iv++) node[iv][ix][iy][iz]= UNKNOWN;
            }

                               /* assign input data to the nearest grid node */

    if (isim==1) fprintf(fperr,"\nClassification of points\n");

    for (i=1;i<=ndata;i++) {
        ix=NINT((coord[i].x-origin.x)/delta.x)+1;
        iy=NINT((coord[i].y-origin.y)/delta.y)+1;
        iz=NINT((coord[i].z-origin.z)/delta.z)+1;
        iv=index[i];

        if( ix>=1 && ix<=n_nodes.x && iy>=1 && iy<=n_nodes.y && iz>=1
                  && iz<=n_nodes.z ) {    /* point is within simulation area */
            if(node[iv][ix][iy][iz]== UNKNOWN) {
                node[iv][ix][iy][iz]=value[i];
                node_flag[ix][iy][iz]=0;
                if((i<=10 || i>= ndata-10) && isim==1) {
                    fprintf(fperr,
            "%8.2f %8.2f %8.2f, %3d, %8.2f assigned to           %4d %4d %4d\n",
                          coord[i].x,coord[i].y,coord[i].z,index[i],
                          value[i],ix,iy,iz);
                    fprintf(fperr,
                    "                                 value ");
                    fprintf(fperr,"%6.3f ",node[iv][ix][iy][iz]);
                    fprintf(fperr,"\n");
                }
            } else if (isim==1) {          /* point has already been assigned */
                fprintf(fperr,
                "%8.2f %8.2f %8.2f, %3d, %8.2f cannot be assigned to %4d %4d %4d\n",
                    coord[i].x,coord[i].y,coord[i].z,index[i],value[i],ix,iy,iz);
                fprintf(fperr,
                "        because another sample is already there\n");
            }
        } else if (isim==1) {              /* point outside simulation area */
            fprintf(fperr,
            "%8.2f %8.2f %8.2f, %3d %8.2f is out of bounds     (%4d %4d %4d)\n",
                          coord[i].x,coord[i].y,coord[i].z,
                          index[i],value[i],ix,iy,iz);
        }
    }
}


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - initialize_covariance_tables()  - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void initialize_covariance_tables(fperr,dbg,ncov,srch_dim,delta,cova_table,
                                    variogram,lookup_table)
FILE        *fperr;
int         ncov,dbg,**lookup_table;
ipt         srch_dim;
pt          delta;
float       ***cova_table[];
variogram_model variogram[];
{

/* This subroutine initializes the covariance tables. There are (nvar*(nvar+1)/2)
   covariance tables. The size of the smallest parallelepiped that
   includes the search neighborhood is passed in the 'ipt' srch_dim. The
   covariances are computed only for positive values of dx, since an
   ergodic covariance is assumed */

    pt  point1,point2;
    int ix,iy,iz,i,j,counter=0;

    printf("Initializing covariance tables. Table:   \n");

    point1.x=0;
    point1.y=0;
    point1.z=0;

    /* covariance values are computed for vectors from grid node (0,0,0)
       to grid node (ix,iy,iz) as (ix,iy,iz) sweeps the search nieghborhood */

    for (i=1;i<=ncov;i++) {
        printf("-------->   %d .. \n",i);
        for (ix=0;ix<=2*srch_dim.x;ix++)
            for (iy= -2*srch_dim.y;iy<=2*srch_dim.y;iy++)
                for(iz= -2*srch_dim.z;iz<=2*srch_dim.z;iz++) {
                    point2.x=ix*delta.x;
                    point2.y=iy*delta.y;
                    point2.z=iz*delta.z;
                    cova_table[i][ix][iy][iz]=
                              cova(point1,point2,&variogram[i]);
                }
    }

    for (i=1;i<=ncov;i++) for(j=1;j<=i;j++) {
         counter++;
         lookup_table[i][j]=lookup_table[j][i]=counter;
    }

    printf("\n\n");

    if (dbg >= 3) {

    /* show tables */

        int i,ix,iy,iz;
        fprintf(fperr,"\n____COVARIANCE TABLES_____\n");
        fprintf(fperr,"Dimensions of the search neigborhood: %d, %d, %d\n",
                        srch_dim.x,srch_dim.y, srch_dim.z);
        for (i=1;i<=ncov;i++) {
            fprintf(fperr,"Table: %d\n",i);
            for (ix=0;ix<=2*srch_dim.x;ix++) {
                fprintf(fperr,"\n ----------------------------> ix=%d\n",ix);
                for(iy= -2*srch_dim.y;iy<=2*srch_dim.y;iy++) {
                    for(iz= -2*srch_dim.z;iz<=2*srch_dim.z;iz++) {
                        fprintf(fperr,"%8.3f",cova_table[i][ix][iy][iz]);
                    }
                    fprintf(fperr,"\n");
                }
            }
        }
    }
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - initialize_template() - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void initialize_template(fperr,dbg, tmp,dim,srch,delta,cova_table)
FILE        *fperr;
int         dbg;
template    *tmp;
ipt         dim;
pt          delta;
float       ***cova_table;
search      srch;
{
/* This subroutine finds the relative position of all the nodes within the
   search nighborhood with respect to the central node. The distance to
   the center and the octant is also obtained. Finally the nodes in the
   template are sorted from closest to farthest to the center. The octants
   are numbered from 0 to 7 */

    int     ix,iy,iz,i,counter;
    pt      h;
    float   norm_h,oct_rot[3][3];

    /* rotation to dealign the cartesian axes to the anisotropy
       axes to make the octant search */

    /* oct_rot[0][0]=oct_rot[0][1]=oct_rot[0][2]=oct_rot[1][2]=oct_rot[2][2]=.58;
    oct_rot[1][0]=oct_rot[2][1]= -0.79;
    oct_rot[1][1]=0.21; oct_rot[2][0]= -0.21;
    */

    printf("Initializing template\n\n");

    counter=0;
    for(ix= -dim.x;ix<=dim.x;ix++)
        for(iy= -dim.y;iy<=dim.y;iy++)
            for(iz= -dim.z;iz<=dim.z;iz++) {
                h.x=ix*delta.x;
                h.y=iy*delta.y;
                h.z=iz*delta.z;
                h=rotate(h,srch.cosines);
                norm_h=h.x/srch.radius.x*h.x/srch.radius.x+
                       h.y/srch.radius.y*h.y/srch.radius.y+
                       h.z/srch.radius.z*h.z/srch.radius.z;
                if(norm_h <1) {
                    /* h=rotate(h,oct_rot); */
                    counter++;
                    tmp->irx[counter]=ix;
                    tmp->iry[counter]=iy;
                    tmp->irz[counter]=iz;
                    if(iz<0) {
                        tmp->octant[counter]=4;
                        if(ix<=0 && iy>0) tmp->octant[counter]=1;
                        if(ix>0 && iy>=0) tmp->octant[counter]=2;
                        if(ix<0 && iy<=0) tmp->octant[counter]=3;
                    } else {
                        tmp->octant[counter]=0;
                        if(ix<=0 && iy>0) tmp->octant[counter]=5;
                        if(ix>0 && iy>=0) tmp->octant[counter]=6;
                        if(ix<0 && iy<=0) tmp->octant[counter]=7;
                    }
 /*                   tmp->octant[counter]=
                            (h.x<0 ? 4 : 0) + (h.y<0 ? 2 : 0) + (h.z<0 ? 1 : 0); */
/*                    tmp->distance[counter]=ix*ix+iy*iy+iz*iz; */
/* now a structural distance is used instead of the euclidean distance used
   before. */
                   tmp->distance[counter]= -lookup_cova(ix,iy,iz,cova_table)
                                           +0.0001*norm_h;
                }
            }

    printf("Sorting template\n\n");

    hpsort(counter,tmp->distance,tmp->index);

    if (dbg >= 3) {         /* show template */
        int j,k;
        fprintf(fperr,"Distance  index  ix  iy  iz  octant\n");
        fprintf(fperr,"-----------------------------------\n");
        for(j=1;j<=counter;j++) {
            k=tmp->index[j];
            fprintf (fperr,"%6f %4d %4d %4d %4d %4d\n",tmp->distance[j],k,tmp->irx[k],
                    tmp->iry[k],tmp->irz[k],tmp->octant[k]);
        }
        fprintf(fperr,"\n");
    }
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - lookup_cova() - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float lookup_cova(dx,dy,dz,cova_table)
int         dx,dy,dz;
float       ***cova_table;
{
/* This function returns the value stored in the covariance table 'cova_table'
   for a distance vector (measured in grid nodes) given by, dx, dy and dz.
   Since the covariance tables were computed for positive values of dx only,
   a radial symmetry is applied if dx<0 */

    if(dx<0) {
        dx= -dx;
        dy= -dy;
        dz= -dz;
    }
    return cova_table[dx][dy][dz];
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - -  a_few_checks() - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void a_few_checks(ndata,ntemp,ncov,nvar,nvar_sim,n_nodes,from,to,delta)
int     ndata, ntemp, ncov,nvar,nvar_sim;
ipt n_nodes,from,to;
pt  delta;
{
/* This function ends the initialization tasks with a few consistency checks
   in the input data. If any of this test is failed the program is aborted */

    int i;

    if(ndata < 0) error("Number of data less than zero");
    if(ntemp <= 0)
        error("Search neighborhood too small, no data points in template");
    if(nvar <= 0) error("Number of variables less than or equal to zero");
    if(nvar_sim > nvar)
         error ("Number of simulated variables larger than number of variables\n");
    if(n_nodes.x<=0 || n_nodes.y<=0 || n_nodes.z<=0)
        error("Error in number of nodes, no nodes in one direction");
    if(from.x>to.x || from.y>to.y || from.z>to.z)
        error("Incosistency between \'from\' and \'to\'");
    if(from.x<=0 || from.x >n_nodes.x || from.y<0 || from.y>n_nodes.y ||
        from.z<=0 || from.z >n_nodes.z) error("\'from\' out of limits");
    if(to.x<=0 || to.x>n_nodes.x || to.y<=0 || to.y>n_nodes.y || to.z<=0 ||
        to.z>n_nodes.z) error("\'to\' out of limits");
    if(delta.x==0 || delta.y==0 || delta.z==0)
        error("Size of grid block equal to zero");

}






/*****************************************************************************/
/*****************************************************************************/
/******************         qcksrt.c                   ***********************/
/*****************************************************************************/
/*****************************************************************************/


/* This is quicksort, the quickest sorting algorithm. From page 251 of
   Numerical Recipes in C by Press et al., Cambridge University
   Press */

#define M 20
#define NSTACK 100
#define FM 7875
#define FA 211
#define FC 1663

/* M is the size of subarrays sorted by straight insertion,
   NSTACK is required auxiliary storage, and the remaining
   constants are used by the random number generating
   statements */

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - qcksrt()  - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void qcksrt(n,arr,indx)
int n, indx[];
float arr[];


/* sorts arr[1...n] into ascending numerical order, indx[1...n]
   is an index array sorted along with arr */

{
    int l=1,jstack=0,j,ir,iq,i,ia;
    int istack[NSTACK+1];
    long int fx=0L;
    float a;
    void error();

    for (j=1;j<=n;j++) indx[j]=j; /* initialize index */

    ir=n;
    for(;;) {
        if (ir-l < M) { /* sort by straight insertion */
            for (j=l+1;j<=ir;j++) {
                a=arr[j];
                ia=indx[j];
                for (i=j-1;arr[i]>a && i>0;i--) {
                    arr[i+1]=arr[i];
                    indx[i+1]=indx[i];
                }
                arr[i+1]=a;
                indx[i+1]=ia;
            }
            if (jstack == 0) return;
            ir=istack[jstack--];
            l=istack[jstack--];

        } else {
            i=l;
            j=ir;
            fx=(fx*FA+FC) % FM; /* generate a random integer between l
                                   and ir inclusive */
            iq=l+((ir-l+1)*fx)/FM;
            a=arr[iq];
            ia=indx[iq];
            arr[iq]=arr[l];
            indx[iq]=indx[l];
            for(;;) {
                while (j > 0 && a < arr[j]) j--;
                if (j <= i) {
                    arr[i]=a;
                    indx[i]=ia;
                    break;
                }
                arr[i]=arr[j];
                indx[i]=indx[j];
                i++;
                while (a > arr[i] && i <= n) i++;
                if (j <= i) {
                    arr[j]=a;
                    indx[j]=ia;
                    i=j;
                    break;
                }
                arr[j]=arr[i];
                indx[j]=indx[i];
                j--;
            }
            if (ir-i >= i-l) {
                istack[++jstack]=i+1;
                istack[++jstack]=ir;
                ir=i-1;
            } else {
                istack[++jstack]=l;
                istack[++jstack]=i-1;
                l=i+1;
            }
            if (jstack > NSTACK) error("NSTACK too small in QCKSRT");
        }
    }
}
#undef M

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - hpsort()  - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

/* Replaces qcksort, that has failed in some very large arrays */


void hpsort(n, ra, indx)
int n;
float ra[];
int indx[];
{ 
int i,ir,j,l,indxt; 
float rra;
int x,y; 

	for(j=1;j<=n;j++) { indx[j]=j; } 
	
	if (n < 2) return; 

	l=(n >> 1)+1; 
	ir=n; 
	for (;;) { 
		if (l > 1) { 
			y=--l; 
			rra=ra[y]; 
			indxt=indx[y]; 
		} 
		else { 
			rra=ra[ir]; 
			indxt=indx[ir]; 
			ra[ir]=ra[1]; 
			indx[ir]=indx[1]; 
			if (--ir == 1) { 
				ra[1]=rra; 
				indx[1]=indxt; break; 
			} 
		} 
		i=l; j=l+l; 
		while (j <= ir) { 
			if (j < ir && ra[j] < ra[j+1]) j++; 
			if (rra < ra[j]) { 
				ra[i]=ra[j]; 
				indx[i]=indx[j]; 
				i=j; 
				j <<= 1; 
			} 
			else j=ir+1; 
		} 
		ra[i]=rra; 
		indx[i]=indxt; 
	} 
}

#define M 714025
#define IA 1366
#define IC 150889

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - rand()  - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float rand0(idum)
long *idum;

/* returns a uniform random deviate between 0.0 and 1.0.
   set idum to a negative value to initialize or reinitialize
   the sequence. This random number generator has period effectively
   infinite. Its principal limitation is that it returns one of
   only 714025 possible values, equally spaced as a comb in the
   interval [0,1). From page 212 of Numerical Recipes in C, Press
   et al., Cambridge University Press, 1988 */

{
    static long iy,ir[98];  /* the number 98, of elements in the shuffle
                               table is uninmportant */
    static int iff=0;
    int j;
    void error();

    if (*idum < 0 || iff == 0) {
        iff=1;
        if ((*idum=(IC-(*idum)) % M) < 0) *idum = -(*idum);
        for (j=1;j<=97;j++) {           /* initialize the shuffle table */
            *idum=(IA*(*idum)+IC) % M;
            ir[j]=(*idum);
        }
        *idum=(IA*(*idum)+IC) % M;
        iy=(*idum);
    }               /* here is where it starts except on initialization */
    j=1 + 97.0*iy/M;
    if(j > 97 || j < 1) error("RAND: This cannot happen");
    iy=ir[j];
    *idum=(IA*(*idum)+IC) % M;
    ir[j]=(*idum);
    return (float) iy/M;
}



/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - read_in_geometry()  - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void read_in_geometry(fp,fperr,seed,nsim,max_data,delta,origin,n_nodes,from,to,
                      srch,max_per_octant,dbg)
long int    *seed;
pt          *delta,*origin;
ipt         *n_nodes,*to,*from;
int         *max_per_octant,*dbg,*nsim,*max_data;
search      *srch;
FILE        *fp, *fperr;
{
/* this function reads in the geometry parameters as described in main() */

    int i,j;

    /* read parameters */

    printf("Reading Geometry\n\n");

    fscanf(fp,"%ld %d",seed,nsim);
    fscanf(fp,"%f %f %f",&delta->x, &delta->y, &delta->z);
    fscanf(fp,"%f %f %f",&origin->x, &origin->y, &origin->z);
    fscanf(fp,"%d %d %d",&n_nodes->x,&n_nodes->y,&n_nodes->z);
    fscanf(fp,"%d %d %d",&from->x, &from->y, &from->z);
    fscanf(fp,"%d %d %d",&to->x, &to->y, &to->z);
    fscanf(fp,"%f %f %f",&srch->radius.x, &srch->radius.y, &srch->radius.z);
    for(i=0;i<3;i++) {
        for(j=0;j<3;j++) {
            fscanf(fp,"%f",&srch->cosines[i][j]);
        }
    }
    srch->rotation=findout_rotation(srch->cosines);
    fscanf(fp,"%d %d %d",max_per_octant,max_data,dbg);
    if (ferror(fp)) error("Error reading in GEOMETRY");

    /* write them out */

    fprintf(fperr,"seed..................:  %ld\n",*seed);
    fprintf(fperr,"number of simulations.:  %d\n",*nsim);
    fprintf(fperr,"grid spacing..........:  %10.3f,%10.3f,%10.3f\n",
                   delta->x,delta->y,delta->z);
    if(delta->x<0 || delta->y<0 || delta->z<0) {
        fprintf(fperr,
            "Negative grid spacing indicates that the grid index increases\n");
        fprintf(fperr,"in the opposite direction as the coordinate does\n");
    }
    fprintf(fperr,"origin................:  %10.3f %10.3f %10.3f\n",
                   origin->x,origin->y,origin->z);
    fprintf(fperr,"domain size...........:  %4d %4d %4d\n",n_nodes->x,
                   n_nodes->y,n_nodes->z);
    fprintf(fperr,"subgrid from..........:  %4d %4d %4d\n",
                   from->x,from->y,from->z);
    fprintf(fperr,"subgrid to............:  %4d %4d %4d\n",to->x,to->y,to->z);
    fprintf(fperr,"search ellipsoid radii:%10.3f %10.3f %10.3f\n",
                   srch->radius.x,srch->radius.y,srch->radius.z);
    fprintf(fperr,"\nDirection cosines for search neighborhood\n");
    for(i=0;i<3;i++) {
        for(j=0;j<3;j++) {
            fprintf(fperr,"%f  ",srch->cosines[i][j]);
        }
        fprintf(fperr,"\n");
    }
    switch(srch->rotation) {
        case 0: fprintf(fperr,"\nno rotation\n"); break;
        case 1: fprintf(fperr,"\nx rotation\n"); break;
        case 2: fprintf(fperr,"\ny rotation\n"); break;
        case 3: fprintf(fperr,"\nz rotation\n"); break;
        case 4: fprintf(fperr,"\ngeneral rotation\n");
    }
    if(*max_per_octant==0)
      fprintf(fperr,"\nno octant search implemented\n");
    else
      fprintf(fperr,"\nmax. point per octant.: %d\n",*max_per_octant);
    fprintf(fperr,  "max. overall data.....: %d\n",*max_data);
    fprintf(fperr,  "debugging flag........: %d\n",*dbg);
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - findout_rotation()  - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

int findout_rotation(cosines)
float cosines[3][3];
{
      int rotation;

      /* no rotation --- identity matrix */

      if(  fabs((double)cosines[0][0]-1.0) < TINY &&
           fabs((double)cosines[1][1]-1.0) < TINY &&
           fabs((double)cosines[2][2]-1.0) < TINY &&
           fabs((double)cosines[0][1])     < TINY &&
           fabs((double)cosines[0][2])     < TINY &&
           fabs((double)cosines[1][0])     < TINY &&
           fabs((double)cosines[1][2])     < TINY &&
           fabs((double)cosines[2][0])     < TINY &&
           fabs((double)cosines[2][1])     < TINY  )  rotation=0;

      /* rotation about the x-axis */

      else if(  fabs((double)cosines[0][0]-1.0) < TINY &&
                fabs((double)cosines[0][1])     < TINY &&
                fabs((double)cosines[0][2])     < TINY &&
                fabs((double)cosines[1][0])     < TINY &&
                fabs((double)cosines[2][0])     < TINY  )  rotation=1;

      /* rotation about the y-axis */

      else if(  fabs((double)cosines[1][1]-1.0) < TINY &&
                fabs((double)cosines[1][0])     < TINY &&
                fabs((double)cosines[1][2])     < TINY &&
                fabs((double)cosines[0][1])     < TINY &&
                fabs((double)cosines[2][1])     < TINY  )  rotation=2;

      /* rotation about the z-axis */

      else if(  fabs((double)cosines[2][2]-1.0) < TINY &&
                fabs((double)cosines[2][0])     < TINY &&
                fabs((double)cosines[2][1])     < TINY &&
                fabs((double)cosines[0][2])     < TINY &&
                fabs((double)cosines[1][2])     < TINY  )  rotation=3;

      else rotation=4;

      return (rotation);
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - read_in_variograms()  - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

int read_in_variograms(fp,fperr,nvar_sim,simulation_var,
                       expected_value,variogram)
FILE *fp,*fperr;
float  *expected_value[];
int    *simulation_var[],*nvar_sim;
variogram_model *variogram[];
{
/* this functions reads the variogram information and returns the number of
   variables considered in this problem. For each variable
   it is required the expected_value value and cross-variograms. */

    int i,nvar;

    printf("Reading Variogram\n\n");

                       /* read in number of variables */

    fscanf(fp,"%d",&nvar);
    fscanf(fp,"%d",nvar_sim);

                                                          /* allocate memory */

    *expected_value=vector(1,nvar);
    *simulation_var=ivector(1,*nvar_sim);

    *variogram=(variogram_model *)malloc((size_t) (nvar*(nvar+1)/2)*
                sizeof(variogram_model));
    if(!*variogram) error("allocation failure in variogram");
    (*variogram)--;  /* this last sentence makes the first element of the
                     vector of variograms have index 1 instead of 0 */

                           /* read simulation_var, expected_value values
                                                              and structures */

    for(i=1;i<=*nvar_sim;i++)
        fscanf(fp,"%d",*simulation_var+i);
    for(i=1;i<=nvar;i++)
        fscanf(fp,"%f",*expected_value+i);
    for(i=1;i<=(nvar*(nvar+1))/2;i++)
        input_structure(fp,*variogram+i);

    if (ferror(fp)) error("Error reading in VARIOGRAM");

                                                     /* write everything out */

    fprintf(fperr,"\nNumber of variables.: %2d\n",nvar);
    fprintf(fperr,  "N. of var. for sim..: %2d --> ",*nvar_sim);
    for(i=1;i<=*nvar_sim;i++) fprintf(fperr,"%2d",*(*simulation_var+i));
    fprintf(fperr,"\n==========================\n");
    fprintf(fperr,"Expected Values\n");
    for(i=1;i<=nvar;i++)
        fprintf(fperr,"Marginal means     :%10.3f\n",*(*expected_value+i));
    for(i=1;i<=(nvar*(nvar+1))/2;i++)
        print_structure(fperr,*variogram+i);
    return nvar;
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - input_structure() - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void input_structure(fp,var_model)
FILE *fp;
variogram_model *var_model;
{
/* this function reads the data corresponding to a variogram model that
   could have several nested structures */

    int i,j;


    fscanf(fp,"%f",&(var_model->nugget));
    fscanf(fp,"%f",&(var_model->cmax));

    fscanf(fp,"%d",&var_model->num_struct);

                              /* allocate memory for the arrays in var_model */

    var_model->type   = ivector(0,var_model->num_struct-1);
    var_model->sill   = vector(0,var_model->num_struct-1);
    var_model->a      = vector(0,var_model->num_struct-1);
    var_model->anis   = (pt *) malloc((size_t) var_model->num_struct*
                                       sizeof(pt));

    /* the old way of reading 'a' and the anisotropy factors has been
       changed to just reading the ranges in the three axes, x', y', z'
       in which the original cartesian axes x, y, z are transformed. To
       modify the minimum this part of the code and the rest of the
       code, first the ranges are read in anis.x, anis.y and anis.z
       then a is made equatl to anis.x and the anis.x, anis.y,
       are recalculated to be old values */


    for(i=0;i<var_model->num_struct;i++) {
        fscanf(fp,"%d",&var_model->type[i]);
    }

    for(i=0;i<var_model->num_struct;i++) {
        fscanf(fp,"%f",&var_model->sill[i]);
    }


    for(i=0;i<var_model->num_struct;i++) {
        fscanf(fp,"%f",&var_model->anis[i].x);
    }

    for(i=0;i<var_model->num_struct;i++){
        fscanf(fp,"%f",&var_model->anis[i].y);
    }

    for(i=0;i<var_model->num_struct;i++){
        fscanf(fp,"%f",&var_model->anis[i].z);
    }

    /* after reading ranges in anis.x, anis.y, anis.z, recompute
       a, anis.x, anis.y and anis.z the old way */

    for(i=0;i<var_model->num_struct;i++) {
        var_model->a[i]=var_model->anis[i].x;
        var_model->anis[i].x=var_model->a[i]/var_model->anis[i].x;
        var_model->anis[i].y=var_model->a[i]/var_model->anis[i].y;
        var_model->anis[i].z=var_model->a[i]/var_model->anis[i].z;
    }

    for(i=0;i<3;i++) {
        for(j=0;j<3;j++) {
            fscanf(fp,"%f",&var_model->dir_cosines[i][j]);
        }
    }

}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - print_structure() - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void print_structure(fp,var_model)
variogram_model *var_model;
FILE *fp;
{
/* this function prints the variogram structure stored in 'var_model' */
    int i,j;


    fprintf(fp,"Nugget.......................: %6.3f\n",var_model->nugget);
    fprintf(fp,"Max Cova (for linear models).: %6.3f\n",var_model->cmax);
    fprintf(fp,"\nNumber of structures --> %d\n\n",var_model->num_struct);
    for(i=0;i<var_model->num_struct;i++)
        switch (var_model->type[i]) {
            case 2: fprintf(fp,
                        "Structure: %d, type...........: Exponential\n",i+1);
                    break;
            case 1: fprintf(fp,
                        "Structure: %d, type...........: Spherical  \n",i+1);
                    break;
            case 3: fprintf(fp,
                        "Structure: %d, type...........: Gaussian   \n",i+1);
                    break;
            case 4: fprintf(fp,
                        "Structure: %d, type...........: Power      \n",i+1);
                    break;
            case 5: fprintf(fp,
                        "Structure: %d, type...........: Power      \n",i+1);
                    break;
            default: error("Variogram type is not defined!!");
        }
    for(i=0;i<var_model->num_struct;i++)
        switch (var_model->type[i])  {
            case 4: fprintf(fp,
             "Structure: %d, scaling .......: %6.3f\n",i+1,var_model->sill[i]);
                    break;
            case 1:
            case 2:
            case 3:
			case 5: fprintf(fp,
             "Structure: %d, sill...........: %6.3f\n",i+1,var_model->sill[i]);
                    break;
        }
    for(i=0;i<var_model->num_struct;i++) {
        fprintf(fp,
           "Structure: %d, range (power for linear model) on rotated x: %6.3f\n",i+1,
                       var_model->a[i]/var_model->anis[i].x);
        fprintf(fp,
           "Structure: %d, range (power for linear model) on rotated y: %6.3f\n",i+1,
                       var_model->a[i]/var_model->anis[i].y);
        fprintf(fp,
           "Structure: %d, range (power for linear model) on rotated z: %6.3f\n",i+1,
                       var_model->a[i]/var_model->anis[i].z);
    } fprintf(fp,"\n");
    fprintf(fp,"Direction cosines\n");
    for(i=0;i<3;i++) {
        for(j=0;j<3;j++) {
            fprintf(fp,"%8.3f",var_model->dir_cosines[i][j]);
        }
        fprintf(fp,"\n");
    }

}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - read_in_conditioning_data() - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

int read_in_conditioning_data(fp,fperr,coord_pointer,index_pointer,
                               value_pointer)
FILE  *fp,*fperr;
float **value_pointer;
int   **index_pointer;
pt    **coord_pointer;
{
/* this function reads in the conditioning data */

    int   i,ndata;
    float *value;
    pt    *coord;
    int   *index;

    printf("Reading hard conditinioning data\n\n");

    fscanf(fp,"%d",&ndata);

                                                          /* allocate memory */
    value=vector(1,ndata);
    index=ivector(1,ndata);
    coord=(pt *)malloc((size_t) ndata*sizeof(pt));
    coord--;  /* this is to set the first element of coord[] to have index 1 */

    for(i=1;i<=ndata;i++)
        fscanf(fp,"%f%f%f%d%f",&(coord[i].x),&(coord[i].y),
                       &(coord[i].z),index+i,value+i);
    if (ferror(fp)) error("Error reading in DATA");

                                      /* write a few of the conditioning data*/

    fprintf(fperr,"\n\nSome conditioning points follow: x, y, z, var and value\n");
    fprintf(fperr,"Number of data: %d\n",ndata);
    for(i=1;i<=ndata;i++)
        if(i<=10 || i>ndata-10)
            fprintf(fperr,"%10.3f,%10.3f,%10.3f,%3d,%10.3f\n",coord[i].x,
                coord[i].y,coord[i].z,index[i],value[i]);

    *value_pointer=value;
    *coord_pointer=coord;
    *index_pointer=index;
    return ndata;
}




/* the following subroutines are from the book Numerical Recipes in C by
   Press, Flannery, Teukolsky and Vetterling, Cambridge University Press,
   1988. They allow to allocate and free vectors, matrices of higher order
   tensors with arbitrary indices. This subroutines overcome the drawback
   of C that makes the first index of all vectors or matrices be 0 */

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - vector()  - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/


float *vector(nl,nh)
int nl,nh;
{
/* returns a pointer to a vector of float numbers with indices
'nl' to 'nh' */

	float *v;

	v=(float *)malloc((size_t)(nh-nl+1)*sizeof(float));

	if(!v) error("allocation failure in vector()");

	return v-nl;
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - ivector()   - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

int *ivector(nl,nh)
int nl,nh;
{
/* returns a pointer to a vector of integer numbers with indices
'nl' to 'nh' */

	int *v;

	v=(int *)malloc((size_t)(nh-nl+1)*sizeof(int));

	if(!v) error("allocation failure in ivector()");

	return v-nl;
}

	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
	/*- - - - - - - - - - - - - lvector()   - - - - - - - - - - - - - - - - - - -*/
	/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

	long *lvector(nl,nh)
	long nl,nh;
	{
	/* returns a pointer to a vector of integer numbers with indices
	'nl' to 'nh' */

		long *v;

		v=(long *)malloc((size_t)(nh-nl+1)*sizeof(long));

		if(!v) error("allocation failure in lvector()");

		return v-nl;
	}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - dvector()   - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

double *dvector(nl,nh)
int nl,nh;
{
/* returns a pointer to a vector of double precision numbers with indices
'nl' to 'nh' */

	double *v;

	v=(double *)malloc((size_t)(nh-nl+1)*sizeof(double));

	if(!v) error("allocation failure in dvector()");

	return v-nl;
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - matrix()  - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float **matrix(nrl,nrh,ncl,nch)
int nrl,nrh,ncl,nch;
{
/* returns a pointer to a matrix of float numbers with first index
'nrl' to 'nrh',and second index 'ncl' to 'nch' */

	int i;
	float **m;

	/* allocate pointers to rows */

	m=(float **) malloc((size_t)(nrh-nrl+1)*sizeof(float *));
	if(!m) error("allocation failure 1 in matrix()");

	m -= nrl;

	/* allocate rows and set pointers to them */

	for (i=nrl;i<=nrh;i++) {
		m[i]=(float *) malloc((size_t)(nch-ncl+1)*sizeof(float));
		if(!m[i]) error("allocation failure 2 in matrix()");
		m[i] -= ncl;
	}

	/* return pointer to array of pointers */

	return m;
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - tensor()  - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float ***tensor(nrl,nrh,ncl,nch,nll,nlh)
int nrl,nrh,ncl,nch,nll,nlh;
{
/* returns a pointer to a tensor of float numbers with first index
'nrl' to 'nrh',and second index 'ncl' to 'nch' and third index 'nll' to
'nlh' */

	int i,j;
	float ***t;

	/* allocate a column of pointers to rows */

	t=(float ***) malloc((size_t) (nrh-nrl+1)*sizeof(float **));
	if(!t) error("allocation failure 1 in tensor()");

	t-=nrl;

	/* allocate rows of pointers and set pointers to them */

	for (i=nrl;i<=nrh;i++) {
		t[i]=(float **) malloc((size_t)(nch-ncl+1)*sizeof(float *));
		if(!t[i]) error("allocation failure 2 in tensor()");
		t[i]-=ncl;
	}

	/* allocate levels and set pointers to them */

	for (i=nrl;i<=nrh;i++) {
		for (j=ncl;j<=nch;j++) {
			t[i][j]=(float *) malloc((size_t)(nlh-nll+1)*sizeof(float));
			if(!t[i][j]) error("allocation failure 3 in tensor()");
			t[i][j]-=nll;
		}
	}
	return t;
}


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - free_vector()   - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void free_vector(v,nl,nh)

float *v;
int nl,nh;
{
/* this function frees the vector of float numbers 'v' previously
   allocated with index from 'nl' to 'nh' */
        v+=nl;
	free (v);
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - free_ivector()  - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void free_ivector(v,nl,nh)

int *v;
int nl,nh;
{
/* this function frees the vector of interger numbers 'v' previously
   allocated with index from 'nl' to 'nh' */

	free ((void *)(v+nl));
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - free_lvector()  - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void free_lvector(v,nl,nh)

long *v;
long nl,nh;
{
/* this function frees the vector of interger numbers 'v' previously
   allocated with index from 'nl' to 'nh' */

	free ((void *)(v+nl));
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - -free_matrix()  - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void free_matrix(m,nrl,nrh,ncl,nch)
float **m;
int nrl,nrh,ncl,nch;
{
/* this function frees the marix of float numbers 'm' previously
   allocated with indices from 'nrl' to 'nrh' and 'ncl' to 'nch' */

	int i;
	for (i=nrh;i>=nrl;i--) {
	    m[i]+=ncl;
	    free(m[i]);
	}

	m+=nrl;
	free((void *)(m));
}
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - imatrix() - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/


int **imatrix(nrl,nrh,ncl,nch)
int nrl,nrh,ncl,nch;
{
	int i,**m;

	m=(int **)malloc((unsigned) (nrh-nrl+1)*sizeof(int*));
	if (!m) error("allocation failure 1 in imatrix()");
	m -= nrl;

	for(i=nrl;i<=nrh;i++) {
		m[i]=(int *)malloc((unsigned) (nch-ncl+1)*sizeof(int));
		if (!m[i]) error("allocation failure 2 in imatrix()");
		m[i] -= ncl;
	}
	return m;
}


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - error() - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void error(err)
/* displays an error message and stops program */
char *err;
{

	while(*err) printf("%c",*err++);
	exit(0);
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - print_header()  - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void print_header(fperr)
FILE  *fperr;
{
	printf(" G C O S I M 3 D version "); printf(VER);printf(DATE);
	printf("\n ======================================\n\n");
	printf(AUTHOR1);printf("\n");printf(ADDRESS1);printf("\n\n");
	printf("\n\ncopyright ");printf(COPYRIGHT);printf("\n\n\n");
	fprintf(fperr," G C O S I M 3 D version "); fprintf(fperr,VER);fprintf(fperr,DATE);
	fprintf(fperr,"\n ====================================\n\n");
	fprintf(fperr,AUTHOR1);fprintf(fperr,"\n");fprintf(fperr,ADDRESS1);
	fprintf(fperr,"\n\ncopyright ");fprintf(fperr,COPYRIGHT);fprintf(fperr,"\n\n\n");

}



